/*
 * Sablotron - XSL processor
 * Copyright (C) 2000 Ginger Alliance Ltd.
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
 * GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#ifndef BaseHIncl
#define BaseHIncl

/****************************************
*                                       *
    i n c l u d e s
*                                       *
****************************************/

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>

#include "shandler.h"

// see also end of file

/****************************************
*                                       *
    c o n s t a n t s
*                                       *
****************************************/

//the number of slash-separated parts of a location path:
#define LOCPATH_SIZE 8
//the number of alternatives (separated by |) in a union of nodesets:
#define LOCPATH_ALTS 8
//the number of predicates in a LocStep:
#define LOC_PRED_COUNT 8
//maximum expression length in chars:
#define MAX_EXPR_LEN 80
//# of tokens in expression:
#define EXPR_NUM_TOKENS 20

//max # of params and args passed to SablotProcessVA
#define ARG_COUNT 32

/*................................................................
    Default sizes for lists and stacks
    (the structures can grow dynamically; this number is the
    size of a 'block' which is initially allocated)
................................................................*/

#define DEFSIZE_CONTEXTLIST 4
#define SCONTEXT_SIZE 16
#define UCONTEXT_DEPTH 8
#define STACK_SIZE_V 8
#define XSTACK_SIZE 16
#define PARSE_BUFSIZE 256
#define RULES_NO 20

#define EPS 1E-10
#define PRIORITY_NOMATCH -10e6

#define NS_SEP '`'


/****************************************
*                                       *
    e n u m s
*                                       *
****************************************/

// vertex type. To be replaced by OTYPE (Sablot object type) (??)

enum VTYPE
{
    VT_VERTEX,VT_ROOT,VT_ELEMENT,VT_ATTRIBUTE,VT_TEXT,
        VT_PI, VT_COMMENT, VT_NAMESPACE,
	  VT_VERTEX_WF = VT_VERTEX,
        VT_ATTRIBUTE_WF = VT_ATTRIBUTE,
        VT_TEXT_WF = VT_TEXT,
        VT_BASE = 0x000f,
        VT_DADDY_FLAG = 0x2000,
        VT_DADDY_WF = VT_VERTEX | VT_DADDY_FLAG,
        VT_ELEMENT_WF = VT_ELEMENT | VT_DADDY_FLAG,  //WF = with flags
        VT_ROOT_WF = VT_ROOT | VT_DADDY_FLAG,
        VT_XSL = 0x4000,
        VT_XSL_ELEMENT_WF = VT_XSL | VT_ELEMENT_WF
};

// SpeakMode. Used to distinguish between vertex output for logging and for 'official' output.

enum SpeakMode
{
    SM_NAME = 1,
        SM_CONTENTS = 2,

    SM_OFFICIAL = SM_NAME | SM_CONTENTS,

        SM_INS_SPACES = 0x1000,
        SM_DESCRIBE = 0x2000,
        SM_ESCAPE = 0x4000
};

/****************************************
    X S L   e n u m s
****************************************/
// the corresponding string tables are declared later and defined in base.cpp

enum	XSL_OP
{
    XSL_APPLY_IMPORTS, XSL_APPLY_TEMPLATES,
        XSL_ATTRIBUTE, XSL_ATTRIBUTE_SET,
        XSL_CALL_TEMPLATE, XSL_CHOOSE,
        XSL_COMMENT, XSL_COPY, XSL_COPY_OF,
        XSL_DECIMAL_FORMAT, XSL_ELEMENT,
        XSL_FALLBACK, XSL_FOR_EACH,
        XSL_IF, XSL_IMPORT,
        XSL_INCLUDE, XSL_KEY,
        XSL_MESSAGE, XSL_NAMESPACE_ALIAS,
        XSL_NUMBER, XSL_OTHERWISE,
        XSL_OUTPUT, XSL_PARAM,
        XSL_PRESERVE_SPACE, XSL_PROCESSING_INSTR,
        XSL_SORT, XSL_STRIP_SPACE,
        XSL_STYLESHEET, XSL_TEMPLATE,
        XSL_TEXT, XSL_TRANSFORM,
        XSL_VALUE_OF, XSL_VARIABLE,
        XSL_WHEN, XSL_WITH_PARAM,
        XSL_NONE
};

enum XSL_ATT
{
    XSLA_CASE_ORDER, XSLA_CDATA_SECT_ELEMS, XSLA_COUNT, 
    XSLA_DATA_TYPE, XSLA_DECIMAL_SEPARATOR, XSLA_DIGIT, XSLA_DISABLE_OUTPUT_ESC, XSLA_DOCTYPE_PUBLIC, XSLA_DOCTYPE_SYSTEM,
    XSLA_ELEMENTS, XSLA_ENCODING, XSLA_EXCL_ELEM_PREFIXES, XSLA_EXT_ELEM_PREFIXES, 
    XSLA_FORMAT, XSLA_FROM, 
    XSLA_GROUPING_SEPARATOR, XSLA_GROUPING_SIZE,
    XSLA_HREF, 
    XSLA_ID, XSLA_INDENT, XSLA_INFINITY, 
    XSLA_LANG, XSLA_LETTER_VALUE, XSLA_LEVEL,
    XSLA_MATCH, XSLA_MEDIA_TYPE, XSLA_METHOD, XSLA_MINUS_SIGN, XSLA_MODE,
    XSLA_NAME, XSLA_NAMESPACE, XSLA_NAN, 
    XSLA_OMIT_XML_DECL, XSLA_ORDER, 
    XSLA_PATTERN_SEPARATOR, XSLA_PERCENT, XSLA_PER_MILLE, XSLA_PRIORITY,
    XSLA_RESULT_PREFIX,
    XSLA_SELECT, XSLA_STANDALONE, XSLA_STYLESHEET_PREFIX, 
    XSLA_TERMINATE, XSLA_TEST,
    XSLA_USE, XSLA_USE_ATTR_SETS, 
    XSLA_VALUE, XSLA_VERSION, 
    XSLA_ZERO_DIGIT,
    XSLA_NONE
};

enum ExAxis
{
    AXIS_ANCESTOR, AXIS_ANC_OR_SELF, AXIS_ATTRIBUTE, AXIS_CHILD,
        AXIS_DESCENDANT, AXIS_DESC_OR_SELF, AXIS_FOLLOWING, AXIS_FOLL_SIBLING,
        AXIS_NAMESPACE, AXIS_PARENT, AXIS_PRECEDING, AXIS_PREC_SIBLING,
        AXIS_SELF,
        AXIS_NONE,
        
        //not recognized in stylesheet:
        AXIS_ROOT
};

enum ExNodeType
{
    EXNODE_NODE,
        EXNODE_TEXT,
        EXNODE_PI,
        EXNODE_COMMENT,
        EXNODE_NONE
};

/****************************************
*                                       *
    t y p e d e f s
*                                       *
****************************************/

enum Bool
{
    FALSE = 0,
        TRUE = 1
};

/****************************************
*                                       *
    m a c r o s
*                                       *
****************************************/

#define NONZERO(x) (assert(x),x)
#define NZ(x) NONZERO(x)
#define cdelete(PTR) {if (PTR) delete PTR; PTR = NULL;}

#ifdef _DEBUG
#define cast(TYPE,PTR) (NONZERO(dynamic_cast<TYPE>(PTR)))
#else
#define cast(TYPE,PTR) ((TYPE) PTR)
#endif

// macros for checking vertex types

#define baseType(v) (NZ(v) -> vt & VT_BASE)
#define isDaddy(v) (NZ(v) -> vt & VT_DADDY_FLAG)
#define isRoot(v) (baseType(v) == VT_ROOT)
#define isElement(v) (baseType(v) == VT_ELEMENT)
#define isText(v) (baseType(v) == VT_TEXT)
#define isXSL(v) (NZ(v) -> vt & VT_XSL)
#define isXSLElement(v) ((isElement(v)) && (isXSL(v)))
#define isAttr(v) (baseType(v) == VT_ATTRIBUTE)
#define isNS(v) (baseType(v) == VT_NAMESPACE)


#define OK FALSE
#define NOT_OK TRUE
#define eFlag Bool
#define BOOL Bool

// to support the change to Unicode
#define Mchar char
#define Mstrlen strlen
#define Mstrchr strchr
#define Mstrcmp strcmp
#define Mstrncmp strncmp

// messaging macro from "Advanced Windows"
// #pragma Msg(some_message) emits a message at compile time
#define MsgToS(x) #x
#define MsgToS2(x) MsgToS(x)
#define Msg(x) message(__FILE__"(" MsgToS2(__LINE__) "): "#x"        <---")


/****************************************
*                                       *
    e x t e r n s
*                                       *
****************************************/

extern char* xslOpNames[];
extern char* xslAttNames[];
extern char* axisNames[];
extern char* exNodeTypeNames[];
extern char* vertexTypeNames[];
extern int lookup(char* str, char** table);
extern int fcomp(double,double);	//float comparison
extern Bool isstd(const char *);
extern int stdclose(FILE *f);
extern FILE *stdopen(const char *fn, const char *mode);
extern Bool strEqNoCase(const char*, const char*);
//
class Processor;
extern Processor *proc;
class SituationObj;
extern SituationObj situation;

extern void pushNewHandler();
extern void popNewHandler();

extern SchemeHandler *theSchemeHandler;

/****************************************
*                                       *
    c l a s s e s
*                                       *
****************************************/

/*
class SabObj
{
public:
    void *operator new(unsigned int);
    void operator delete(void *, unsigned int);
};
*/

/*****************************************************************
Memory leak test stuff
*****************************************************************/
#if defined(CHECK_LEAKS)
#pragma Msg("Building with leak checking")
#ifdef WIN32
#define CRTDBG_MAP_ALLOC 
#include <crtdbg.h>
#elif defined(__linux__)
#include <mcheck.h>
#endif
#endif

extern void checkLeak();
extern void memStats();


/****************************************
*                                       *
    f i n a l   i n c l u d e s
*                                       *
****************************************/

#ifndef BaseH_NoErrorH
#include "situa.h"
#else
#undef BaseH_NoErrorH
#endif

#endif //ifndef BaseHIncl

