/*
 * Sablotron - XSL processor
 * Copyright (C) 2000 Ginger Alliance Ltd.
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
 * GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

// datastr.h

#ifndef DataStrHIncl
#define DataStrHIncl

#define LIST_SIZE 16
#define STACK_DEPTH 16

// error.h will be included later:
#define BaseH_NoErrorH
#include "base.h"
#undef BaseH_NoErrorH


/****************************************
L i s t
****************************************/

template <class T>
class List /* : public SabObj */
{
public:
    List(int = LIST_SIZE);
    ~List(void);
    T *data;    //array of T's
    int number() const;
    void append(T x);
    T& operator[](int ndx) const;
    T& last();
    void deppend();
    void deppendall();
    void rm(int);
    Bool isEmpty();
    void swap(int,int);
protected:
    int nItems; 
private:
    int blocksize;
    int currBlock, nBlocks;
    struct ListBlock 
    {
        T *data;
        ListBlock *prev, *next;
    } *block0, *blockn;
    void *appendBlock();
    void deppendBlock();
};

/****************************************
P L i s t
a list of pointers
****************************************/

template <class T>
class PList : public List<T>
{
public:
    PList(int = LIST_SIZE);
    void freelast(Bool);
    void freeall(Bool);
    void freerm(int, Bool);
};

/****************************************
S t a c k
****************************************/

template <class T>
class Stack : public List<T>
{
public:
    Stack(int = STACK_DEPTH);
    void push(T what);
    T pop();
    T top();
    void reset();
};

/****************************************
S t r
****************************************/

class Str
{
public:
    Str();
    Str(char);
    Str(char *);
    Str(const Str&);
    ~Str();
    operator char*();
    Str& operator= (const Str &);
    Str& operator= (char*);
    Str& operator+= (const char*);
    Str& operator= (char);
    Str& operator+= (char);
    Str& operator= (int);
    Str& operator= (double);
    Str& operator+= (int);
    Str operator+ (const char *);
    Str operator+ (Str&);
    Str operator+ (int);
    Bool operator== (const char*);
    Bool operator== (char*);
    Bool operator== (Str&);
    Bool eqNoCase(const char*);
    Bool operator< (char*);
    Bool operator< (Str &);
    int compare(char *);
    char operator[] (int);
    int length();
    Str& nset(const char *,int);
    Str& add(char *);
    Str& nadd(const char *,int);
    BOOL isEmpty() const;
    Bool toDouble(double&);
    void empty();
    void speakTerse(Str &);
    static Bool namechar(char c);
    static Bool isdigit_(char c);
    void forget();
    void remove();
private:
    void compact();
//    char *_string;
    int _length;
    PList<char*> _parts;
};

extern Str theEmptyString;

/****************************************
Q N a m e
****************************************/
// container for the QName (e.g. books:isbn) consising
// of the "namespace prefix" and the "local part"

class NSList;

class QName /* : public SabObj */
{
public:
    void speak(Str&,SpeakMode) const;
    eFlag setFromExpat(const Str &);
    eFlag setLogical(const Str&, NSList *, Bool);
    void setPrefix(const Str&);
    Bool prefixIsNS(Str &s);
    Bool operator==(QName &);
    Str getname();
    const Str& getLocal() const;
    const Str& getUri() const;
    void findPrefix(const NSList&);
    Str 
        prefix,
        uri,
        local;
};

/*****************************************************************
    S t r S t r L i s t
*****************************************************************/

class StrStr /* : public SabObj */
{
public:
    Str 
        key,
        value;
};

class StrStrList : public PList<StrStr*>
{
public:
    int findNum(const Str &_key);
    Str* find(Str &);
};

/*****************************************************************

    S L i s t

*****************************************************************/

template <class T>
class SList: public PList<T>
{
public:
    SList(int);
    virtual int compare(int, int) = 0;
    void insert(T);
};

/*****************************************************************

t e m p l a t e   m e t h o d s

*****************************************************************/

#include "situa.h"
//#include "error.h"

/*================================================================
List methods
================================================================*/

template <class T>
List<T>::List(int ablocksize) 
//ablocksize = LIST_SIZE
:
blocksize(ablocksize), nItems(0), 
nBlocks(0), block0(NULL), blockn(NULL)
{};


template <class T>
void List<T>::append(T what)
{
    int ndxInBlock = nItems % blocksize;
    if (!ndxInBlock)
            ((ListBlock*)appendBlock()) -> data = new T[blocksize];
    blockn -> data[ndxInBlock] = what;
    nItems++;
};

template <class T>
void List<T>::deppend()
{
    assert (nItems > 0);
    int ndxInBlock = (nItems-1) % blocksize;
    T item = blockn -> data[ndxInBlock];
    if (!(--nItems % blocksize))
        deppendBlock();
};

template <class T>
void List<T>::rm(int n)
{
    assert((n >= 0) && (n < nItems));
    for (int i = n; i < nItems - 1; i++)
        (*this)[i] = (*this)[i+1];
    deppend();
}

template <class T>
void List<T>::swap(int i, int j)
{
    assert((i >= 0) && (i < nItems));
    assert((j >= 0) && (j < nItems));
    T& p = (*this)[i];
    T& q = (*this)[j];
    T temp = p;
    p = q;
    q = temp;
}

template <class T>
void List<T>::deppendall() 
{
    while(nItems > 0) deppend();
}

template <class T>
int List<T>::number() const
{
    return nItems;
}

template <class T>
List<T>::~List()
{
    deppendall();
};

template <class T>
T& List<T>::operator[](int ndx) const
{
    int x = 0; ////
    assert((ndx < nItems) && (ndx >= 0));
    BOOL back = (ndx >= nItems / 2) ? TRUE : FALSE;
    int ndxInBlock = ndx % blocksize,
        blockNdx = ndx / blocksize,
        steps = (back? nBlocks - blockNdx - 1: blockNdx),
        i;
    ListBlock *b = (back? blockn : block0);
    for (i = 0; i < steps; i++)
        b = (back? b -> prev : b -> next);
    return b -> data[ndxInBlock];
};

template <class T>
T& List<T>::last()
{
    assert(nItems);
    return (*this)[nItems - 1];
}

template <class T>
void* List<T>::appendBlock()
{
    ListBlock *b = new ListBlock;
    if (nBlocks) 
        blockn -> next = b;
    else
        block0 = b;
    b -> prev = blockn;
    blockn = b;
    b -> next = NULL;
    nBlocks++;
    return b;
}

template <class T>
void List<T>::deppendBlock()
{
    assert(nBlocks);
    ListBlock *b = blockn;
    blockn = b -> prev;
    delete[] b-> data;
    delete b;
    nBlocks--;
    if (!nBlocks)
        block0 = NULL;
}

template <class T>
Bool List<T>::isEmpty()
{
    return (Bool) (nItems == 0);
}

/*================================================================
PList methods
================================================================*/

template <class T>
PList<T>::PList(int ablocksize)
: List<T>(ablocksize)
{
}
 
template <class T>
void PList<T>::freelast(Bool asArray)
{
    if (asArray)
        delete[] last();
    else
        delete last();
    deppend();
}

template <class T>
void PList<T>::freeall(Bool asArray)
{
    while (nItems > 0)
        freelast(asArray);
}

template <class T>
void PList<T>::freerm(int n, Bool asArray)
{
    assert((n >= 0) && (n < nItems));
    if (asArray)
        delete [] (*this)[n];
    else
        delete (*this)[n];
    rm(n);
}

/*================================================================
Stack methods
================================================================*/

template <class T>
Stack<T>::Stack(int ablocksize) 
// ablocksize = STACK_DEPTH
:
List<T>(ablocksize)
{};

template <class T>
void Stack<T>::push(T what)
{
    append(what);
}

template <class T>
T Stack<T>::pop()    
{
    T retval = (*this)[nItems-1];
    deppend();
    return retval;
}

template <class T>
T Stack<T>::top(void)
{
    return (*this)[nItems-1];
}

template <class T>
void Stack<T>::reset(void)
{
    deppendall();
}

/*================================================================
SList methods    
================================================================*/

template <class T>
SList<T>::SList(int anum) 
    : PList<T>(anum)
{
};

template <class T>
void SList<T>::insert(T what)
{
    append(what);
    int whatndx = number() - 1;
    int i, j;
    for (i=0; i < whatndx; i++)
        if (compare(whatndx,i) == -1) break;
    if (i < whatndx)
    {
        for (j = whatndx; j > i; j--)
            operator[](j) = operator[](j-1);
        operator[](i) = what;
    };
};

#endif //ifndef DataStrHIncl

