/*
 * Sablotron - XSL processor
 * Copyright (C) 2000 Ginger Alliance Ltd.
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
 * GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include "base.h"
#include "proc.h"
#include "xmlparse.h"
#include "expr.h"
#include "tree.h"
#include "context.h"
#include "vars.h"

/*
  include direct.h for getcwd()
*/
#if defined(__linux__)
#include <unistd.h>
#elif defined(WIN32)
#include <direct.h>
#elif defined(__BORLANDC__)
#include <dir.h>
#endif

/*
  the XSLT namespace
*/

Str ns_XSL = "http://www.w3.org/1999/XSL/Transform";


/*****************************************************************
R u l e I t e m   methods
*****************************************************************/

RuleItem::RuleItem(XSLElement *arule,double aprio, QName &_name,
                   QName *_mode)
:
rule(arule), priority(aprio), name(_name), mode(_mode)
{
};


//
//  ~RuleItem
//  has to dispose of the mode
//
RuleItem::~RuleItem()
{
    cdelete(mode);
}


/*****************************************************************
R u l e S L i s t   methods
*****************************************************************/

RuleSList::RuleSList()
:
SList<RuleItem*>(RULES_NO)
{};

RuleSList::~RuleSList()
{
    freeall(FALSE);
}

int RuleSList::compare(int i, int j)
{
    return fcomp(operator[](i) -> priority,
        operator[](j) -> priority);
};


XSLElement *RuleSList::findByName(QName &what)
{
    for (int i=0; i < number(); i++)
        if (proc -> cmpQNames((*this)[i] -> name, what))
            return (*this)[i] -> rule;
    return NULL;
}

/*****************************************************************

  DataLineItem

*****************************************************************/

DataLineItem::~DataLineItem()
{
    // the following close should be checked for error!!
    if (_dataline)
        _dataline -> close();
    cdelete(_dataline);
    cdelete(_tree);
}

/*****************************************************************

  DataLinesList

*****************************************************************/

int DataLinesList::findNum(Str &absoluteURI, Bool _isXSL,
    DLAccessMode _mode)
{
    for (int i = 0; i < number(); i++)
    {
        DataLineItem* item = (*this)[i];
        if ((item->_dataline -> fullUri == absoluteURI) &&
            (item->_isXSL == _isXSL) && 
            (item-> _dataline -> mode == _mode))
        return i;
    }
    return -1;
}

Tree *DataLinesList::getTree(Str &absoluteURI, Bool _isXSL, 
    DLAccessMode _mode)
{
    int n;
    if ((n = findNum(absoluteURI, _isXSL, _mode)) != -1)
        return (*this)[n] -> _tree;
    else
        return NULL;
}

eFlag DataLinesList::addLine(DataLine *d, Tree *t, Bool isXSL)
{
    DataLineItem *item = new DataLineItem;
    item -> _dataline = d;
    item -> _tree = t;
    item -> _isXSL = isXSL;
    append(item);
    return OK;
}

/*****************************************************************
*                                                                *
*   P r o c e s s o r   methods                                  *
*                                                                *
*****************************************************************/

Processor::Processor()
{
    pushNewHandler();
    
    vars = new VarsList;
    input = styleSheet = result = NULL;
    outTree = NULL;
    hardBaseURI = NULL;

    allocObjects = allocBytes = 
        maxAllocObjects = maxAllocBytes = 0;
};

Processor::~Processor()
{
/*
These deletes are now subsumed in "datalines.freeall()":
    cdelete(input);
    cdelete(result);
    cdelete(styleSheet);
*/
    cdelete(vars);
    datalines.freeall(FALSE);
    argList.freeall(FALSE);
    if (!situation.isError())
        assert(modes.isEmpty());
    situation.clear();
    Log1(L_STOP, (char*) situation.timeStr());
    situation.closeFiles();
    popNewHandler();
};

void my_getcwd(Str &dir)
{
char buf[256];
#if defined(WIN32)
    _getcwd(buf, 256);
#else //if defined(__linux__) || defined(__BORLANDC__)
    getcwd(buf,256);
#endif
    dir = buf;
    if (!(dir == "/"))
        dir += '/';
}

eFlag Processor::open(char *sheetURI, char *inputURI, char *resultURI)
{
    Str temp,
        theBase;
    my_getcwd(theBase);
    theBase = Str("file://") + theBase;
    E( readTreeFromURI(styleSheet, temp = sheetURI, theBase, TRUE) );
    E( readTreeFromURI(input, temp = inputURI, theBase, FALSE) );
    E( makeTreeForURI(result, temp = resultURI, theBase) );
    outTree = result;

    Context *c = new Context;
    c -> set(&(input -> root));

    /*
        in vars, global prebindings go to call level 0,
        global bindings to call level 1
    */
    vars -> startCall();
    E( styleSheet -> root.execute(c) );
    vars -> endCall();

    delete c;
    return OK;
}


/*================================================================
getVarBinding 
    finds the current binding for the variable with the given name.
    The binding is an expression. If none is present, returns NULL.
================================================================*/

Expression *Processor::getVarBinding(QName &which)
{
    return vars -> getBinding(which);
}

/*================================================================
cmpQNames
    compares two given QNames based on the namespace URIs,
    allowing wildcard * as local name of first QName.
================================================================*/

Bool Processor::cmpQNames(QName& first, QName &second)
{
    if (first.local == (const char *) "*")
        return (Bool)(first.prefix.isEmpty() || 
            (first.uri == second.uri));
    else
        return (Bool) (first == second);
}

/*================================================================
aliasXL
    given a namespace prefix, looks at the alias list. If it finds
    an entry, translates the prefix; otherwise returns the same
    prefix.
================================================================*/

Str& Processor::aliasXL(Str& prefix)
{
    Str *p;
    if (p = aliases.find(prefix))
        return *p;
    else
        return prefix;
}

/*================================================================
execute
================================================================*/

eFlag Processor::execute(Vertex *IP, Context *c)
{
    while (!c -> isFinished())
    {
        if (IP)
            E( IP -> execute(c) )
        else
            E( execApply(c) );
        c -> shift();
    };
    delete c;
    return OK;
}

eFlag Processor::execute(VertexList &IPlist, Context *c)
{
    while (c -> current())
    {
        E( IPlist.execute(c) );
        c -> shift();
    };
    delete c;
    return OK;
}

eFlag Processor::execApply(Context *c)
{
    XSLElement *rule;
    E( bestRule(rule, c) );
    if (!rule)
        E( builtinRule(c) )
    else
        E( rule -> execute(c) );
    return OK;
}

/*================================================================
bestRule
    Finds the highest-priority rule that is satisfied by v. 
    Assumes that 'rules' is sorted by priority in reverse order.
    If several rules are satisfied, returns the one that occurs last. 
    If no rule is found, returns NULL.
================================================================*/

eFlag Processor::bestRule(XSLElement *& ret, Context *c)
{
    int i;
    int bestRule = -1;
    double bestPrio=0;
    Expression *pattern;
    QName 
        *currMode = proc -> getCurrentMode(),
        *thisMode;
    XSLElement *therule;
    for (i = 0; i < rules.number(); i++)
    {
        if (
            (bestRule != -1) &&
            (fcomp(rules[i] -> priority, bestPrio) == -1)  //current < best
            )
            break;
        Attribute *a = (therule = rules[i] -> rule) -> atts.find(XSLA_MATCH);
        if (a) 
            pattern = a -> expr;
        else 
            continue;

        thisMode = rules[i] -> mode;
        if ((!thisMode) ^ (!currMode))
            continue;
        if (thisMode && !((*thisMode) == (*currMode)))
            continue;
        // else both thisMode and currMode are NULL which is OK
        if (pattern && (pattern -> matches(c)))
        {            
            bestRule = i;
            bestPrio = rules[i] -> priority;
        };
    };
    if (bestRule == -1) ret = NULL;
    else ret = rules[bestRule] -> rule;
    return OK;
};

eFlag Processor::builtinRule(Context *c)
{
    Vertex *v = c -> current();
    
    switch(v -> vt & VT_BASE)
    {
    case VT_ROOT:
    case VT_ELEMENT:
        //apply-templates
        {
            Expression *e = new Expression(NULL,EXF_LOCPATH);
            e -> setLS(
                AXIS_CHILD, 
                EXNODE_NODE);
            Context *newc;
            E( e -> createContext(newc = c) );
            E( execute(NULL, newc) );
            delete e; ///
        }; break;
    case VT_TEXT:
    case VT_ATTRIBUTE:
        //copy contents to output
        {
            Str temp;
            E( v -> value(temp, c) );
            getOutTree() -> appendVertex(new Text(temp));
        }; break;
    case VT_PI:
    case VT_COMMENT:
        {};
    };
    return OK;
};

eFlag Processor::insertRule(XSLElement *tmpl)
{
    double prio;
    Attribute *a = tmpl -> atts.find(XSLA_PRIORITY);
    if (!a)
        prio = defaultPriority(tmpl);
    else
    {
        if (a -> cont.toDouble(prio))
            Err(ET_BAD_NUMBER);
    };
    QName q, *mode = NULL;
    if (!!(a = tmpl -> atts.find(XSLA_NAME)))
        E( q.setLogical(a -> cont, &(tmpl -> namespaces), FALSE) );
    if (!q.local.isEmpty() && rules.findByName(q))
        Err1(ET_DUPLICATE_RULE_NAME,(char*)(q.getname()));
    if (!!(a = tmpl -> atts.find(XSLA_MODE)))
        E( (mode=new QName) -> setLogical(a -> cont, 
            &(tmpl -> namespaces), FALSE) );
    
    rules.insert(new RuleItem(tmpl,prio,q,mode));
    return OK;
}

double Processor::defaultPriorityLP(Expression *lpath)
{
    assert(lpath && lpath -> functor == EXF_LOCPATH);
    assert(lpath -> args.number());
    if (lpath -> args.number() > 1)
        return .5;
    else
    {
        switch (lpath -> args[0] -> step -> ntype)
        {
        case EXNODE_COMMENT:
        case EXNODE_TEXT:
        case EXNODE_PI:  //// pi('literal') should have priority 0.5
        case EXNODE_NODE:
            return -0.5;
            break;
        case EXNODE_NONE:
            {
                QName &qn = lpath -> args[0] -> step -> ntest;
                if (!(qn.local == (const char *) "*"))
                    return 0.0;
                else
                {   
                    if (qn.prefix.isEmpty())
                        return -0.5;
                    else
                        return -0.25;
                };
            }; break;
        default:
            return 0.5;
        };
    };
    return 0;   // BCC thinks we don't return enough
}

double Processor::defaultPriority(XSLElement *tmpl)
{
    Expression *e = tmpl -> getAttExpr(XSLA_MATCH);
    if (!e) 
        return PRIORITY_NOMATCH;
    switch(e -> functor)
    {
    case EXF_LOCPATH:
        {
            return defaultPriorityLP(e);
        }; break;
    case EXFO_UNION:
        {
            double max=0, priority;
            BOOL first = TRUE;
            for (int i=0; i < e -> args.number(); i++)
            {
                priority = defaultPriorityLP(e -> args[i]);
                if (first || (priority > max)) 
                    max = priority;
                first = FALSE;
            };
            return max;
        }; break;
    default:
        {
            assert(!"expression not a union or LP");
            return 0;   // dummy
        }
    };
    return 0;       // dummy for BCC
}

Tree *Processor::getOutTree()
{
    return outTree;
}

void Processor::setOutTree(Tree *_outTree)
{
    outTree = _outTree;
}

/*................................................................
getCurrentMode()
RETURN: pointer to the current mode which is a QName
................................................................*/

QName *Processor::getCurrentMode()
{
    return (modes.number() ? modes.last() : NULL);
}

/*................................................................
pushMode()
called before every mode-changing apply-templates
ARGS:
m       the mode to be placed on top of the mode stack
................................................................*/

void Processor::pushMode(QName *m)
{
    modes.append(m);
}

/*................................................................
popMode()
called after every mode-changing apply-templates
removes the current mode from the mode stack FREEING it
................................................................*/

void Processor::popMode()
{
    modes.freelast(FALSE);
}

eFlag Processor::addLineParse(Tree *& newTree, Str &absolute, Bool isXSL)
{
    DataLine *d = new DataLine;
    E( d -> open(absolute, DLMODE_READ) );
    newTree = new Tree(absolute, isXSL);
    newTree -> dline = d;
    E( parse(newTree, d));
    E( datalines.addLine(d, newTree, isXSL) );
    return OK;
}

eFlag Processor::readTreeFromURI(Tree*& newTree, Str& location, Str& base,
      Bool isXSL)
{
    Str
        absolute;
    makeAbsoluteURI(location, base, absolute);
    newTree = datalines.getTree(absolute, isXSL, DLMODE_READ);
    if (!newTree)
        E( addLineParse(newTree, absolute, isXSL) );
    return OK;
}

eFlag Processor::makeTreeForURI(Tree*& newTree, Str& location, Str& base)
{
    Str absolute;
    makeAbsoluteURI(location, base, absolute);
    if (datalines.getTree(absolute, FALSE, DLMODE_WRITE))
        Err1(E1_CANNOT_WRITE, absolute);
    newTree = new Tree(absolute, FALSE);
    DataLine *d = new DataLine;
    newTree -> dline = d;
    E( d -> open(absolute, DLMODE_WRITE) );
    E( datalines.addLine(d, newTree, FALSE) );
    return OK;
}

/*****************************************************************
findStylesheet()
finds a xsl:stylesheet child of the given daddy. Returns NULL
if not found.
*****************************************************************/

XSLElement *findStylesheet(Daddy &d)
{
    Vertex *w;
    for (int i = 0; i < d.contents.number(); i++)
    {
        if (isXSLElement(w = d.contents[i]) && 
            (toX(w) -> op == XSL_STYLESHEET || toX(w) -> op == XSL_TRANSFORM))
            return toX(w);
    };
    return NULL;
}

/*****************************************************************
Processor::processVertexAfterParse()
Performs any necessary actions on a vertex immediately after it is
parsed into the tree. The vertex is still 'current'. Typically, this
function calls popVertex.
ARGS:
    v   the vertex to be processed
    t   the containing tree
The function operates only on XSL vertices, namely:
xsl:include - replaces the vertex by the newly constructed tree
...
Other vertices are just popped off the stack.
*****************************************************************/

eFlag Processor::processVertexAfterParse(Vertex *v, Tree *t)
{
    XSL_OP theOp;
    if (isXSLElement(v))
    {
        XSLElement *x = toX(v);
        switch(theOp = x -> op)
        {
        case XSL_INCLUDE:
            {
                Attribute *a = NZ( x -> atts.find(XSLA_HREF) );
                Tree *srcTree;
                E( readTreeFromURI(srcTree, a -> cont, t -> name,TRUE) );
                
                XSLElement *theSheet = findStylesheet(srcTree -> root);
                if (!theSheet)
                    Warn1(W_NO_STYLESHEET, (char*)(a -> cont));
                t -> dropCurrentElement(v);
                if (theSheet)
                    theSheet -> contents.copy(t, TRUE);
            }; break;
        default:
            t -> popVertex();
        }
    }
    else
        t -> popVertex();
    return OK;
}


/*****************************************************************
*                                                                *
*   Element and text handlers for Expat.                         *
*                                                                *
*****************************************************************/

/*================================================================
hlrStartE
    Element start handler.
================================================================*/

void hlrStartE
(void *parser,const char *elName,const char **atts)
{
    char **p = (char**) atts;
    Tree *t = (Tree*) XML_GetUserData((XML_Parser*)parser);
    XSL_OP opCode;
    XSL_ATT attCode;
    BOOL itsXSL = FALSE;
    Vertex *v;
    Attribute *a;
    QName q;
    int elemLine;

    if (situation.isError())
        return;

    situation.setCurrLine(
        elemLine = XML_GetCurrentLineNumber((XML_Parser*) parser));
    if (q.setFromExpat((char *)elName))
        return;
    if ((t -> XSLTree) && (q.uri == ns_XSL))
    {
        itsXSL = TRUE;
        opCode = (XSL_OP) lookup((char*) q.local,xslOpNames);
        if (opCode == XSL_NONE)
        {
            situation.error(ET_BAD_XSL, theEmptyString, theEmptyString);
            return;
        };
        v = new XSLElement(q, t, opCode);
    }
    else
        v = new Element(q,t);
    v -> lineno = elemLine;
    t -> appendVertex(v);
    
    t -> pendingNS.giveCurrent(toE(v) -> namespaces, t);
    v -> name.findPrefix(toE(v) -> namespaces);
    
    while(*p)
    {
        if (q.setFromExpat((char *)p[0]))
            return;
        q.findPrefix(toE(v) -> namespaces);
        attCode = (itsXSL ? 
            (XSL_ATT) lookup((char*)q.local,xslAttNames) : XSLA_NONE);
        a = new Attribute(q,p[1],attCode);
        a -> lineno = XML_GetCurrentLineNumber((XML_Parser*) parser);
        t -> appendVertex(a);
        p += 2;
    };
    
      if (itsXSL)
          cast(XSLElement*,v) -> checkAtts(); 
      else
          if (t -> XSLTree)
            for (int k = 0; k < cast(Element*,v) -> atts.number(); k++)
                cast(Attribute*, cast(Element*,v) -> atts[k]) -> buildExpr(
                    TRUE, EX_NONE);
        // this calls situation.error() on error
}

/*================================================================
hlrEndE
    Element end handler.
================================================================*/

void hlrEndE(void *parser,const char *elName)
{
    if (situation.isError())
        return;
    Tree *t = (Tree*) XML_GetUserData((XML_Parser*)parser);
    Vertex *v = NZ( t -> stackTop );
    
    if (t -> XSLTree)
        t -> stripped += 
        (isXSLElement(v)? cast(XSLElement*, v) : cast(Daddy*, v))
            -> strip();

    if(isXSLElement(v))
    {
        // situation.error() is called in the following
        if (toX(v) -> checkChildren())
            return;
        if (toX(v) -> processChildren())
            return;
    }
    if (proc -> processVertexAfterParse(v, t))
        return;
    // t -> popVertex();    this is done in processVeretxAfterParse
}

/*================================================================
hlrText
    Text node handler.
================================================================*/

void hlrText(void *parser, const char *s, int len)
{
    if (situation.isError())
        return;
    Vertex *newVertex;
    Tree *t = (Tree*) XML_GetUserData((XML_Parser*)parser);
    if (!!(newVertex = t -> appendText((char *) s,len)))
        newVertex -> lineno = XML_GetCurrentLineNumber((XML_Parser*) parser);
}

/*================================================================
hlrStartNS
    namespace scope start handler
================================================================*/

void hlrStartNS(void *parser, const char *prefix, const char *uri)
{
    if (situation.isError())
        return;
    Tree *t = (Tree*) XML_GetUserData((XML_Parser*)parser);
    /*
    Str tempPfx = prefix? (char *) prefix : "",
      tempUri = (char*) uri;
    */
    Vertex *newv;
    t -> pendingNS.append(newv = new NmSpace(
        prefix? (char *) prefix : (char*)"",
        (char*) uri));
    newv -> lineno = XML_GetCurrentLineNumber((XML_Parser*) parser);
}

/*================================================================
hlrEndNS
    namespace scope end handler
================================================================*/

void hlrEndNS(void *parser, const char *prefix)
{
    if (situation.isError())
        return;
    Tree *t = (Tree*) XML_GetUserData((XML_Parser*)parser);
    // hope this works
    assert(toNS(t -> pendingNS.last()) -> prefix ==
        (prefix ? (char*)prefix : ""));
    t -> pendingNS.freelast(FALSE);
}

/*================================================================
hlrUnkEnc
    unknown encoding handler
================================================================*/

#include "enc1250.h"

int hlrUnkEnc(void *encodingHandlerData, const char *name,
               XML_Encoding *info)
{
    if (!strEqNoCase((char*) name,"windows-1250"))
        return 0;
    int i;
    for (i = 0; i < 0x80; i++)
    {
        info -> map[i] = i;
        info -> map[i + 0x80] = Enc1250[i];
    }
    info -> map[0x7f] = -1;
    info -> data = NULL;
    info -> convert = NULL;
    info -> release = NULL;
    return 1;
}

/*================================================================
Processor::parse
    Uses Expat to parse in a tree from the given source.
================================================================*/

eFlag Processor::parse(Tree *t, DataLine *d)
{
    Bool quit = FALSE;
    int res;
    int bytes;
    char buf[PARSE_BUFSIZE];
    //
    XML_Parser parser = XML_ParserCreateNS(NULL,NS_SEP);
    if (!parser)
        Err(E_MEMORY);
    XML_SetUserData(parser,t);
    XML_UseParserAsHandlerArg(parser);
    XML_SetElementHandler(parser,hlrStartE,hlrEndE);
    XML_SetCharacterDataHandler(parser,hlrText);
    XML_SetNamespaceDeclHandler(parser, hlrStartNS, hlrEndNS);
    XML_SetUnknownEncodingHandler(parser, hlrUnkEnc, NULL);

    situation.pushCurrent();
    situation.setCurrFile(t -> name);
    //
    while (!quit)
    {
        bytes = d -> get(buf,sizeof(buf));
        if (situation.isError())
        {
            XML_ParserFree(parser);
            return NOT_OK;
        };
        quit = (Bool) (bytes < sizeof(buf));
        //        if (!XML_Parse(parser,buf,bytes,quit)) 
        res = XML_Parse(parser,buf,bytes,quit);
	if (situation.isError())
	  return NOT_OK;
	if (!res) 
            {
                situation.setCurrFile(t -> name);
                situation.setCurrLine(XML_GetCurrentLineNumber(parser));
                int code = XML_GetErrorCode(parser); 
                Str eCodeStr, eNameStr;
                eCodeStr = code;
                eNameStr = (char*) XML_ErrorString(code);
                Err2(E_XML, eCodeStr, eNameStr);
                XML_ParserFree(parser);
            }
    };
    if (t -> XSLTree)
        t -> stripped += t -> root.strip();
    situation.popCurrent();
    XML_ParserFree(parser);
    return OK;
};

eFlag Processor::getArg(const char* name, char*& buffer, Bool isUTF16)
{
    Str temp, *value = argList.find(temp = (char*)name);
    if (!value)
        Err1(E1_ARG_NOT_FOUND,(char*) name);
    buffer = (char*) *value;
    return OK;
}

SchemeHandler *Processor::getSchemeHandler()
{
    return theSchemeHandler;
}

/*****************************************************************
copyArg
  called if the result document's location uses the arg:
  scheme. Returns information about the associated named buffer.
  if not found, returns -1 in argOrdinal and NULL in newCopy.
ARGS
  argName       the name of the arg
RETURNS
  *argOrdinal    the ordinal number of this arg. This is the number
                of the call to useArg() which defined the arg.
  newCopy       pointer to a new copy of the arg (allocated via
                malloc())
*****************************************************************/

void Processor::copyArg(const Str& argName, int* argOrdinal,
    char*& newCopy)
{
    Str absolute;
    int lineNo;
    if ((makeAbsoluteURI((Str&)argName, "arg:/", absolute) != URI_ARG)
        || (lineNo = datalines.findNum(absolute, FALSE, DLMODE_WRITE)) == -1)
    {
        newCopy = NULL;
        *argOrdinal = -1;
        return;
    }
    DynBlock *block = NZ( datalines[lineNo] -> _dataline -> getOutBuffer() );
    newCopy = (char*)malloc(block -> byteCount + 1);
    memcpy(newCopy, block -> getPointer(), block -> byteCount);
    newCopy[block -> byteCount] = 0;

    //  set *argOrdinal
    *argOrdinal = argList.findNum((char *)absolute + 4);    // skip 'arg:'

}

eFlag Processor::useArg(char *name, char *val)
{
    assert(name);
    Str nameStr;
    if (*name != '/')
        nameStr = "/";
    nameStr += name;
    if (argList.find(nameStr))
        Err1(E1_DUPLICATE_ARG, nameStr);
    StrStr *p = new StrStr;
    p -> key = nameStr;
    p -> value = val;
    argList.append(p);
    return OK;
}

eFlag Processor::useGlobalParam(char *name, char *val)
{
    assert(name);
    QName q;
    q.local = name;
    Expression *expr = new Expression(NULL,EXF_ATOM);
    expr -> setAtom(val);
    vars -> addPrebinding(q, expr);
    return OK;
}
