/*
 * Sablotron - XSL processor
 * Copyright (C) 2000 Ginger Alliance Ltd.
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
 * GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

/****************************************
*                                       *
i n c l u d e s
*                                       *
****************************************/

#include<iostream.h>
#include <stdarg.h>

#define SablotAsExport
#include"sablot.h"

#include "base.h"
#include "verts.h"
#include "tree.h"
#include "expr.h"
#include "proc.h"
#include "situa.h"

/****************************************
g l o b a l s
****************************************/
Processor *proc = NULL;
SituationObj situation;
Str *keptBuffer = NULL;

#define checkErr() {if (situation.isError()) return NOT_OK;}
#define Guarded(x) try { x; } catch(ErrCode e){ Err(e); }


/****************************************
d e f i n i t i o n s
****************************************/

void doStart()
{
    situation.msgOutputFile("stderr","sablot.log");
    Log1(L_START, situation.timeStr());
};

void doEnd()
{
    Log1(L_STOP, situation.timeStr());
}

#define ErrQ(e) {situation.error(e,theEmptyString,theEmptyString); doEnd(); return NOT_OK;}


/*****************************************************************
SablotProcess_internal()
The main internal interfacing function. Not visible from the outside
so it can evolve as needed without changes to the interface. The other
functions are wrappers around this one.
*****************************************************************/

int SablotProcess_internal(char *sheetURI, char *inputURI, char *resultURI,
    char **params, char **arguments, char **resultArg, char *hardBaseURI)
{
    // to debug memory leaks, uncomment the call to checkLeak()
    // and set _crtBreakAlloc to the # of block (proc=51)
#if defined(__linux__) && defined(CHECK_LEAKS)
    mtrace();
#endif
    Bool problem = FALSE;
    checkErr();
    doStart();
    if (proc)
        ErrQ(E_IN_PROGRESS);
    proc = new Processor;
    if (!proc)
        ErrQ(E_MEMORY);
#ifndef __BORLANDC__
    try
#endif
    {
        char **p;
        if (arguments)
        {
            for (p = arguments; *p && !problem; p += 2)
                problem = proc -> useArg(p[0], p[1]);
        }
        if (params)
        {
            for (p = params; *p && !problem; p += 2)
                problem = proc -> useGlobalParam(p[0], p[1]);
        }
        if (problem || proc -> open(sheetURI,inputURI,resultURI))
        {
            cdelete(proc);
            situation.clear();
            return NOT_OK;
        }
    }
#ifndef __BORLANDC__
    catch (MsgCode e)
    {
        ErrQ(e);
    };
#endif
    proc -> result -> output();

    /* get a copy of the result if it's arg: scheme */
    char *newCopy;
    // we even obtain the index of the output buffer but won't use it
    int _resultNdx;
    proc -> copyArg(resultURI, &_resultNdx, newCopy);
    if (resultArg)
        // if output does not go to a buffer, newCopy is NULL
        *resultArg = newCopy;
    cdelete(proc);
    situation.clear();
    doEnd();
#if defined(WIN32) && defined(CHECK_LEAKS)
    memStats();
    checkLeak();
#endif
    return OK;
};





/*****************************************************************
SablotProcess

  the main function published by Sablotron. Feeds given XML
  input to a stylesheet. Both of these as well as the location for
  output are identified by a URI. One can also pass top-level
  stylesheet parameters and named buffers ('args').
ARGS
  sheetURI      URI of the XSLT stylesheet
  inputURI      URI of the XML document
  resultURI     URI of the output document
  params        a NULL-terminated array of char*'s defining the top-level
                parameters to be passed, interpreted as a
                sequence of (name, value) pairs. 
  arguments     a NULL-terminated array of char*'s defining the named
                buffers to be passed, interpreted as a
                sequence of (name, value) pairs. Both params and arguments
                may be NULL. 
RETURNS
  .             nonzero iff error
  resultArg     in case output goes to a named buffer, *resultArg is set to                     
                point to it (otherwise to NULL). Free it using SablotFree().
*****************************************************************/


int SablotProcess(char *sheetURI, char *inputURI, char *resultURI,
    char **params, char **arguments, char **resultArg)
{
    E( SablotProcess_internal(sheetURI, inputURI, resultURI,
        params, arguments, resultArg, NULL) );
    return OK;
};


/*****************************************************************
SablotProcessVA

  like SablotProcess, but the params and args are given using the ellipsis
  notation (variable arguments). Cannot be used for outputting
  to an arg.

  The optional arguments look like "param1Name", "param1Value",
  "param2Name", "param2Value", ..., NULL, "arg1Name", "arg1Value",
  ..., NULL. 
*****************************************************************/

int SablotProcessVA(char *sheetURI, char *inputURI, char *resultURI, ...)
{
    char *paramsToPass[ARG_COUNT*2+2], *argsToPass[ARG_COUNT*2+2];
    va_list var;
    va_start(var, resultURI);
    int i;

    /*
      read the NULL-terminated params array from the varargs
    */

    for (i = 0; i < ARG_COUNT; i++)
    {
        paramsToPass[2 * i + 1] = NULL;
        if (!!(paramsToPass[2 * i] = va_arg(var, char*)))
            paramsToPass[2 * i + 1] = va_arg(var, char*);
        else break;
    }
    paramsToPass[2*i] = paramsToPass[2 * i + 1] = NULL;

    //    skip any extra params until the separating NULL
    if (i == ARG_COUNT)
        while(va_arg(var, char*));

    //    read the NULL-terminated args array from the varargs

    for (i = 0; i < ARG_COUNT; i++)
    {
        if (!!(argsToPass[2 * i] = va_arg(var, char*)))
            argsToPass[2 * i + 1] = va_arg(var, char*);
        else break;
    }
    argsToPass[2*i] = argsToPass[2 * i + 1] = NULL;

    va_end(var);
    E( SablotProcess_internal(sheetURI, inputURI, resultURI,
        paramsToPass, argsToPass, NULL, NULL) );
    return OK;
}


/*****************************************************************
SablotProcessFiles

  calls SablotProcess to process files identified by their
  file names. No named buffers or params are passed. Included
  for backward compatibility.
ARGUMENTS
  styleSheetName, inputName, resultName
            file names of the stylesheet, XML input and output,
            respectively.
RETURNS
  .         error flag
*****************************************************************/


int SablotProcessFiles(char *styleSheetName,
                       char *inputName,
                       char *resultName)
{
    return SablotProcess_internal(styleSheetName, inputName, resultName, NULL, NULL, NULL, NULL);
};


/*****************************************************************
SablotProcessStrings

  calls SablotProcess to process documents in memory, passing
  pointers to the documents. No named buffers or params are passed.
  Included for backward compatibility.
ARGUMENTS
  styleSheetStr, inputStr
                text of the stylesheet and the XML input
RETURNS
  .             error flag
  *resultStr    pointer to the newly allocated block containing
                the result 
*****************************************************************/
int SablotProcessStrings(char *styleSheetStr,
                         char *inputStr,
                         char **resultStr)
{
  char *argums[] =
  {
    "/_stylesheet", styleSheetStr,
    "/_xmlinput", inputStr,
    "/_output", NULL, //was: *resultStr,
    NULL
  };
  E( SablotProcess_internal("arg:/_stylesheet","arg:/_xmlinput","arg:/_output",
		    NULL, argums, resultStr,
            NULL ) );
  // *resultStr = argums[5];
  return OK;
};


/*****************************************************************
SablotProcessStringsWithBase

    Like SablotProcessStrings but lets you pass an URI that replaces
    the stylesheet's URI in relative address resolution.

ARGUMENTS
  styleSheetStr, inputStr
                text of the stylesheet and the XML input
  theHardBase   the "hard base URI" replacing the stylesheet's URI
                in all relevant situations
RETURNS
  .             error flag
  *resultStr    pointer to the newly allocated block containing
                the result 
*****************************************************************/

int SablotProcessStringsWithBase(char *styleSheetStr,
                         char *inputStr,
                         char **resultStr,
                         char *theHardBase)
{
  char *argums[] =
  {
    "/_stylesheet", styleSheetStr,
    "/_xmlinput", inputStr,
    "/_output", NULL, //was: *resultStr,
    NULL
  };
  E( SablotProcess_internal("arg:/_stylesheet","arg:/_xmlinput","arg:/_output",
		    NULL, argums, resultStr,
            theHardBase ) );
  return OK;
};


/*****************************************************************
SablotFree

  Frees the Sablotron-allocated buffer.
*****************************************************************/

int SablotFree(char *resultStr)
{
    if (resultStr)
        free(resultStr);
    return OK;
}

/*****************************************************************
SablotFreeBuffer

  Not used anymore, kept for compatibility.
*****************************************************************/

int SablotFreeBuffer(char *resultStr)
{
    return OK;
}


/*****************************************************************
SablotRegSchemeHandler

  Registers an external scheme handler. Doing so when a handler is 
  already registered results in a warning.
*****************************************************************/

int SablotRegSchemeHandler(SchemeHandler *sh)
{
    if (theSchemeHandler)
    {
        Warn(W_SH_REGISTERED)
        return NOT_OK;
    };
    theSchemeHandler = sh;
    return OK;
}

/*****************************************************************
SablotUnregSchemeHandler

  Unregisters an external scheme handler. Doing so when no handler is 
  registered results in a warning.
*****************************************************************/

int SablotUnregSchemeHandler()
{
    if (!theSchemeHandler)
    {
        Warn(W_SH_NOT_REGISTERED);
        return NOT_OK;
    }
    theSchemeHandler = NULL;
    return OK;
}

/*****************************************************************
  some retired functions
*****************************************************************/

/*
int SablotGetBuffer(char** value)
{
    checkErr();
    *value = proc -> getOutBuffer();
    return OK;
}

int SablotFreeBuffer(char *resultStr)
{
    checkErr();
    cdelete(proc);
    return OK;
}
*/
