/*
 * Sablotron - XSL processor
 * Copyright (C) 2000 Ginger Alliance Ltd.
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
 * GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

/*****************************************************************
sablot.h
TK Dec 14, 99
Last changed Apr 20, 00
header file for Sablot.cpp
*****************************************************************/

#ifndef SablotHIncl
#define SablotHIncl

#include "shandler.h"

#if defined(__linux__) || defined(__BORLANDC__)
#define DllImpExp extern
#else //__linux__ || __BORLANDC__
#ifdef SablotAsExport
#define DllImpExp __declspec( dllexport )
#else //ifdef SablotAsExport
#define DllImpExp __declspec( dllimport )
#endif //SablotAsExport
#endif //__linux__  || __BORLANDC__


/*****************************************************************
SablotProcess

  the main function published by Sablotron. Feeds given XML
  input to a stylesheet. Both of these as well as the location for
  output are identified by a URI. One can also pass top-level
  stylesheet parameters and named buffers ('args').
ARGS
  sheetURI      URI of the XSLT stylesheet
  inputURI      URI of the XML document
  resultURI     URI of the output document
  params        a NULL-terminated array of char*'s defining the top-level
                parameters to be passed, interpreted as a
                sequence of (name, value) pairs. 
  arguments     a NULL-terminated array of char*'s defining the named
                buffers to be passed, interpreted as a
                sequence of (name, value) pairs. Both params and arguments
                may be NULL. 
RETURNS
  .             nonzero iff error
  resultArg     in case output goes to a named buffer, *resultArg is set to                     
                point to it (otherwise to NULL). Free it using SablotFree().
*****************************************************************/


#ifdef __cplusplus
extern "C" {
#endif
    DllImpExp

    int SablotProcess(char *sheetURI, char *inputURI, char *resultURI,
        char **params, char **arguments, char **resultArg);

#ifdef __cplusplus
}
#endif


/*****************************************************************
SablotProcessFiles

  calls SablotProcess to process files identified by their
  file names. No named buffers or params are passed. Included
  for backward compatibility.
ARGUMENTS
  styleSheetName, inputName, resultName
            file names of the stylesheet, XML input and output,
            respectively.
RETURNS
  .         error flag
*****************************************************************/

#ifdef __cplusplus
extern "C" {
#endif
    DllImpExp

        int SablotProcessFiles(char *styleSheetName,
        char *inputName,
        char *resultName);

#ifdef __cplusplus
}
#endif


/*****************************************************************
SablotProcessStrings

  calls SablotProcess to process documents in memory, passing
  pointers to the documents. No named buffers or params are passed.
  Included for backward compatibility.
ARGUMENTS
  styleSheetStr, inputStr
                text of the stylesheet and the XML input
RETURNS
  .             error flag
  *resultStr    pointer to the newly allocated block containing
                the result 
*****************************************************************/

#ifdef __cplusplus
extern "C" {
#endif

    DllImpExp
        int SablotProcessStrings(char *styleSheetStr,
        char *inputStr,
        char **resultStr);

#ifdef __cplusplus
}
#endif

/*****************************************************************
SablotProcessStringsWithBase

    Like SablotProcessStrings but lets you pass an URI that replaces
    the stylesheet's URI in relative address resolution.

ARGUMENTS
  styleSheetStr, inputStr
                text of the stylesheet and the XML input
  theHardBase   the "hard base URI" replacing the stylesheet's URI
                in all relevant situations
RETURNS
  .             error flag
  *resultStr    pointer to the newly allocated block containing
                the result 
*****************************************************************/

#ifdef __cplusplus
extern "C" {
#endif
    DllImpExp

int SablotProcessStringsWithBase(char *styleSheetStr,
                         char *inputStr,
                         char **resultStr,
                         char *theHardBase);
#ifdef __cplusplus
}
#endif


/*****************************************************************
SablotFree

  Frees the Sablotron-allocated buffer. The user cannot free it
  directly by free().
*****************************************************************/

#ifdef __cplusplus
extern "C" {
#endif
    
    DllImpExp
        int SablotFree(char *resultStr);

#ifdef __cplusplus
}
#endif

/*****************************************************************
SablotFreeBuffer

  Not used anymore. Kept for compatibility.
*****************************************************************/

#ifdef __cplusplus
extern "C" {
#endif
    
    DllImpExp
        int SablotFreeBuffer(char *resultStr);

#ifdef __cplusplus
}
#endif

/*****************************************************************
SablotProcessVA

  like SablotProcess, but the params and args are given using the ellipsis
  notation (variable arguments). Cannot be used for outputting
  to an arg.

  The optional arguments look like "param1Name", "param1Value",
  "param2Name", "param2Value", ..., NULL, "arg1Name", "arg1Value",
  ..., NULL. 
*****************************************************************/

#ifdef __cplusplus
extern "C" {
#endif
    DllImpExp
        int SablotProcessVA(char *sheetURI, 
    	char *inputURI, 
	    char *resultURI, 
    	...);
#ifdef __cplusplus
}
#endif

/*****************************************************************
SablotRegSchemeHandler

  Registers an external scheme handler. Doing so when a handler is 
  already registered results in error.
*****************************************************************/

#ifdef __cplusplus
extern "C" {
#endif
    DllImpExp
    int SablotRegSchemeHandler(SchemeHandler *sh);
#ifdef __cplusplus
}
#endif


/*****************************************************************
SablotUnregSchemeHandler

  Unregisters an external scheme handler. Doing so when no handler is 
  registered results in error.
*****************************************************************/

#ifdef __cplusplus
extern "C" {
#endif
    DllImpExp
    int SablotUnregSchemeHandler();
#ifdef __cplusplus
}
#endif


/*****************************************************************
OUT OF USE FUNCTIONS:

int SablotGetBuffer(char **value)
*****************************************************************/

#endif //defined SablotHIncl

