/*
 * Sablotron - XSL processor
 * Copyright (C) 2000 Ginger Alliance Ltd.
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
 * GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include "base.h"
#include "situa.h"
#include "verts.h"
#include <time.h>
#ifdef SABLOT_CHERROR
#include "chError.h"
#endif

/*****************************************************************

    S   i   t   u   a   t   i   o   n   O   b   j 

*****************************************************************/

SituationObj::SituationObj()
{
    clear();
    logfile = errwfile = NULL;
    msgOutputFile("stderr",NULL);
}

SituationObj::~SituationObj()
{
    if (logfile)
        stdclose(logfile);
    if (errwfile)
        stdclose(errwfile);
}

Str SituationObj::timeStr()
{
    time_t currtime;
    time(&currtime);
    return asctime(localtime(&currtime));
}

void SituationObj::pushCurrent()
{
    SitData *p = new SitData;
    p -> currV = currV;
    p -> currLine = currLine;
    p -> currFile = currFile;
    //history.append(*p);
    history.append(p);
}

void SituationObj::popCurrent()
{
    //SitData &d = history.last();
    SitData *p = history.last();
    currV = p -> currV;
    currLine = p -> currLine;
    currFile = p -> currFile;
    // history.deppend();
    // the history item should be deleted here!!!
    history.freelast(FALSE);
}

void SituationObj::pushCurrV(Vertex *v)
{
    pushCurrent();
    setCurrV(v);
}

void SituationObj::setCurrV(Vertex *v)
{
    currV = v;
    currLine = v -> lineno;
}

void SituationObj::setCurrLine(int lno)
{
    currLine = lno;
}

void SituationObj::setCurrFile(Str& fname)
{
    currFile = fname;
};

eFlag SituationObj::closeFiles()
{
    if (logfile)
        stdclose(logfile);
    logfile = NULL;
    if (errwfile)
        stdclose(errwfile);
    errwfile = NULL;
    return OK;
}

eFlag SituationObj::eraseLog(char *newLogFile)
{
    if (logfile)
        stdclose(logfile);
    logfile = NULL;
    if (newLogFile)
    {
        if (!(logfile = stdopen(newLogFile,"w")))
          Err1(E_FILE_OPEN, newLogFile);
    }
    return OK;
}

eFlag SituationObj::msgOutputFile
(char *_errwfn, char *_logfn)
{
    E( closeFiles() );
    if (_logfn)
    {
        if (!(logfile = stdopen(_logfn,"a")))
          Err1(E_FILE_OPEN, _logfn);
    }
    if (_errwfn)
    {
        if (!(errwfile = stdopen(_errwfn,"w")))
          Err1(E_FILE_OPEN, _errwfn);
    }
    useChError = FALSE;
    return OK;
}

void SituationObj::msgOutputChError()
{
#ifdef CH_ERROR
    E( closeFiles() );
    useChError = TRUE;
#else
    Warn(W_NO_CHERROR);
#endif
}

void SituationObj::message(MsgType type, MsgCode code, 
                         const Str& arg1, const Str& arg2)
{
    char buf[256];
    PList<Str*> out;
    sprintf(buf,"[code:%d] ",code);
    out.append(new Str(buf));
#ifdef SABLOT_CHERROR
    out.append(new Str("[module:Sablotron] "));
#endif
    if (!currFile.isEmpty())
    {
        sprintf(buf,"[URI:%s] ",(char*) currFile);
        out.append(new Str(buf));
    }
    if (currLine)
    {
        sprintf(buf,"[line:%d] ",currLine);
        out.append(new Str(buf));
    }
    if (currV)
    {
        Str nameStr;
        currV -> speak(nameStr, SM_NAME);
        sprintf(buf,"[node:%s%s'%s']",
            vertexTypeNames[currV -> vt & VT_BASE],
            (currV -> vt == VT_VERTEX ? "" : " "),
            (char *) nameStr);
        out.append(new Str(buf));
    }
    int num = !(arg1.isEmpty()) + !arg2.isEmpty();
    SabMsg *p = GetMessage(code);

    if (type != MT_LOG)
        out.append(new Str("\n  "));

    sprintf(buf,p -> text,(char*)(Str&)arg1,(char*)(Str&)arg2);
    out.append(new Str(buf));
    // 'out' could now be passed to the modified chError

    Str fullout;
    if (type != MT_LOG)
    {
        fullout = GetMessage((MsgCode)(MSG_ERROR + type)) -> text;
        fullout += " ";
        for (int j = 0; j < out.number(); j++)
            fullout += *out[j];
    }
    else
        fullout = *(out.last());
    
    if (useChError)
    {
#ifdef SABLOT_CHERROR
        if (type == MT_ERROR)
        {
            chChainError(
                chMakeError(1,0,code),
                (char*) fullout);
        }
        else
        {
            fprintf(stderr,"%s\n", GetMessage(W_CHERROR_ONLY_ERR) -> text);
            fprintf(stderr,"%s\n",(char*) fullout);
        }
#endif
    }
    else
    {
        FILE *thefile = (type == MT_LOG ? logfile : errwfile);
        if (thefile)
            fprintf(thefile,"%s\n",(char*) fullout);
    };
    out.freeall(FALSE);
};

void SituationObj::error(MsgCode code, 
                         const Str& arg1, const Str& arg2)
{
    message(MT_ERROR, code, arg1, arg2);
    message(MT_LOG, code, arg1, arg2);
    pending = code;
}

void SituationObj::warning(MsgCode code, 
                         const Str& arg1, const Str& arg2)
{
    message(MT_WARN, code, arg1, arg2);
    message(MT_LOG, code, arg1, arg2);
}

void SituationObj::logmsg(MsgCode code, 
                         const Str& arg1, const Str& arg2)
{
    message(MT_LOG, code, arg1, arg2);
}

Bool SituationObj::isError()
{
    return (Bool) (pending != E_OK);
};

void SituationObj::clear()
{
    currV = NULL;
    currLine = 0;
    currFile.empty();
    pending = E_OK;
}
