/*
 * Sablotron - XSL processor
 * Copyright (C) 2000 Ginger Alliance Ltd.
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
 * GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

/*****************************************************************
    uri.cpp
*****************************************************************/

#include "uri.h"
#include <string.h>
#include "proc.h"

/*****************************************************************

    global functions

*****************************************************************/

#define RF(CONDITION) {if (!(CONDITION)) return;}

// definition of names for various URI-reference parts
#define U_SCHEME    0
#define U_AUTH      1
#define U_PATH      2
#define U_QUERY     3
#define U_FRAG      4

// definition of slahes in path names
#define slashes "/\\"
#define isSlash(c) (c == '/' || c == '\\')

/*****************************************************************
splitBy

  splits a given string into two parts divided by the first occurence
  of a delimiter from a given set. If no delimiter is found, returns FALSE
  and leaves 'string' as is; otherwise shifts 'string' to the character
  following the delimiter.
ARGS:
  string        the asciiz string to be split
  delims        the asciiz set of delimiters (all of them ASCII chars)
  part1         first of the two parts
RETURN:
  string        shifted to the other part (past the delimiter)
  .             the delimiter found (or 0)
*****************************************************************/

char splitBy(const char *&string, const char *delims, Str &part1)
{
    char c;
    int firstLen = strcspn(string, delims);
    part1.nset(string, firstLen);
    if (!!(c = string[firstLen]))
        string += firstLen + 1;
    return c;
}

typedef Str FiveStr[5];

void splitURI(const char *uri, FiveStr &parts)
{
    const char *rest;
    char c;
    for (int i = 0; i < 5; i++) 
        parts[i].empty();
    RF( *uri );
    // extract the scheme part of the URI
    if (!splitBy(rest = uri, ":", parts[U_SCHEME]))
        parts[U_SCHEME].empty();
    // if "//" follows, extract the authority part
    c = 'A';    // marks the absence of auth
    if (isSlash(*rest) && isSlash(rest[1]))
        RF( c = splitBy(rest += 2, slashes"?#", parts[U_AUTH]) );
    if (isSlash(c) || c == 'A')
        // extract the path
        RF( c = splitBy(rest -= (isSlash(c)), "?#", parts[U_PATH]) );
    if (c == '?')
        // extract the query
        RF( c = splitBy(rest, "#", parts[U_QUERY]) );
    // copy the fragment
    parts[U_FRAG] = (char *) rest;
};

void joinURI(Str &joined, FiveStr &parts, Bool schemeToo)
{
    joined.empty();
    if (schemeToo && !parts[U_SCHEME].isEmpty())
        joined = parts[U_SCHEME] + ":";
    if (!parts[U_AUTH].isEmpty())
        joined += Str("//") + parts[U_AUTH];            // add authority
    joined += parts[U_PATH];                // add path
    if (!parts[U_QUERY].isEmpty())          // add query
        joined += Str("?") + parts[U_QUERY];
    if (!parts[U_FRAG].isEmpty())           // add fragment
        joined += Str("#") + parts[U_FRAG];
}

/*****************************************************************
schemeToURI_()

  converts the scheme given as Str to one of the URI_... constants.
  If the scheme is neither "file" or "arg" then URI_EXTENSION is
  simply returned.
*****************************************************************/

URIScheme schemeToURI_(Str& scheme)
{
    if (scheme.eqNoCase("file"))
        return URI_FILE;
    else
    {
        if (scheme.eqNoCase("arg"))
            return URI_ARG;
        else 
            return URI_EXTENSION;
    }
}

/*****************************************************************
cutLast()

  truncates a path after 'howmany'-th slash from the right (1-based).
  If there are fewer slashes, sets path to empty string and returns
  FALSE, otherwise returns TRUE.
ARGS
  path      the path to be truncated
  howmany   # of slashes that disappear in truncation, MINUS 1
RETURNS
  .         TRUE iff that many slashes were found
  path      the truncated path
*****************************************************************/

Bool cutLast(Str& path, int howmany)
{
    Str temp = path;
    char *p = (char*) temp;
    int slashCount = 0,
        i;
    for (i = temp.length() - 1; i >= 0; i--)
    {
        if (p[i] == '/')
            slashCount++;
        if (slashCount == howmany)
            break;
    };
    if (i >= 0)
        path.nset(p, i+1);
    else
        path.empty();
    return (Bool)(i >= 0);
};

/*****************************************************************
joinPaths()

  merges a relative path with a base path
ARGS
  relPath       the relative path. The result is returned here.
  basePath      the base path (always absolute)
RETURNS
  relPath       the newly constructed absolute path
*****************************************************************/

Bool segP(Str &s, int oneOrTwo)
{
    return (Bool) !strcmp((char *) s, (oneOrTwo == 1 ? "." : ".."));
}

void joinPaths(Str& relPath, const Str& basePath)
{
    Str segment, absPath;
    // append the relPath to all-but-the-last-segment-of-basePath

    Bool endSlash = cutLast(absPath = basePath, 1),
        lastSeg;
    relPath = absPath + (endSlash? "" : "/") + relPath;
       
    // throw out all '.' from the path
    const char *p = (const char*) relPath;
    absPath.empty();
    while(splitBy(p, slashes, segment))
    {
        if (!segP(segment, 1))
            absPath += segment + "/";
    }
    if (!segP(segment, 1))
        absPath += segment;

    // throw out all "something/.." from the path
    p = (char*) absPath;
    int depth = 0;
    relPath.empty();
    do
    {
        lastSeg = (Bool) !splitBy(p, slashes, segment);
        if (!segP(segment, 2))
        {
            relPath += segment + (lastSeg ? "" : "/");
            depth++;
        }
        else
        {
            if (depth > 1)
            {
                cutLast(relPath, 2);
                depth--;
            }
            else
                relPath += segment + (lastSeg ? "" : "/");
        };
    }
    while(!lastSeg);
}


URIScheme makeAbsoluteURI2(const char* uri ,const char* base, Str& absolute, Str& scheme)
{
    FiveStr 
        u_parts, 
        b_parts;
    Bool 
        u_defined[5],
        u_any = FALSE;

    // first, break up the URIs into their 5 components
    splitURI(uri, u_parts);
    splitURI(base, b_parts);

    // set u_defined[i] to TRUE if the i-th uri component is nonvoid
    for (int i = 0; i < 5; i++)
        u_any = (Bool) ((u_defined[i] = (Bool) !u_parts[i].isEmpty()) || u_any);

    if (!u_any) // all components empty: the reference is to the current document
    {
        splitURI(base,u_parts);
        u_parts[U_QUERY].empty();       // query and fragment are NOT inherited from base
        u_parts[U_FRAG].empty();
    }
    else    // not all components are empty
    {
        if (!u_defined[U_SCHEME])                       // undefined scheme
        {
            u_parts[U_SCHEME] = b_parts[U_SCHEME];      // inherit scheme from base
            if (!u_defined[U_AUTH])                     // undefined authority
            {
                u_parts[U_AUTH] = b_parts[U_AUTH];      // inherit authority from base
                if (!isSlash(u_parts[U_PATH][0]))       // path is relative
                    joinPaths(u_parts[U_PATH], b_parts[U_PATH]);    // append path to base path
                // query and fragment stay as they are in 'uri'
            }
        }
        else    // scheme defined, check for paths not starting with '/'
        {
            if (!u_defined[U_AUTH] && !isSlash(u_parts[U_PATH][0]))
                u_parts[U_PATH] = Str("/") + u_parts[U_PATH];
        }
    }
    joinURI(absolute, u_parts, FALSE);         // join all components into a URI for return (no scheme)
    return schemeToURI_(scheme = u_parts[U_SCHEME]);
}


//    URIScheme makeAbsoluteURI(uri, base, absolute)
//
//    Merges a (possibly relative) URI reference with a base URI, setting
//    'absolute' to the result. 
//
URIScheme makeAbsoluteURI(const char* uri ,const char* base, Str& absolute)
{
    Str scheme;
    URIScheme temp;
    temp = makeAbsoluteURI2(uri, base, absolute, scheme);
    absolute = scheme + ":" + absolute;
    return temp;
}


URIScheme uri2SchemePath(const char *absolute, Str& scheme, Str& rest)
{
    Bool found = (Bool) !!splitBy(absolute, ":", scheme);
    assert(found);
    rest = (char*) absolute;
/*
 *    if (isSlash(*absolute) && isSlash(absolute[1]))
 *       rest = (char*) absolute + 2;
 *   else
 *       rest = (char*) absolute;
 */
    return schemeToURI_(scheme);
}


/*****************************************************************
DynBlock

  is a class for "dynamic memory blocks", i.e. blocks that can 
  grow in size.
*****************************************************************/

DynBlock::DynBlock()
{
    first = last = NULL;
    byteCount = 0;
}

DynBlock::~DynBlock()
{
    DynBlockItem *d, *d_was;
    if (first)
    {
        for (d = first; d; )
        {
            d_was = d;
            d = d_was -> next;
            delete d_was -> data;
            delete d_was;
        }
    }
}

void DynBlock::nadd(const char *data, int bytes)
{
    DynBlockItem *newitem = new DynBlockItem;
    newitem -> data = new char[bytes];
    memcpy(newitem -> data, data, bytes);
    newitem -> byteCount = bytes;
    newitem -> next = NULL;
    if (last)
        last -> next = newitem;
    else
        first = newitem;
    last = newitem;
    byteCount += bytes;
}

char *DynBlock::getPointer()
{
    if (!first)
        return NULL;
    if (first -> next)
        compact();
    return (first -> data);
}

void DynBlock::compact()
{
    if (!first || !(first -> next)) return;
    char *newdata = new char[byteCount],
        *p = newdata;
    for (DynBlockItem *d = first; d; d = d -> next)
    {
        memcpy(p, d -> data, d -> byteCount);
        p += d -> byteCount;
        delete[] d -> data;
        delete d;
    };
    first = last = new DynBlockItem;
    first -> data = newdata;
    first -> byteCount = byteCount;
    first -> next = NULL;
}


/*****************************************************************
DataLine

  is a class that holds the machinery needed to retrieve data from
  a given URI. There are two internally supported URI schemes:
  file  (the plain "file://...")
  arg   (for access to named memory blocks passed to Sablotron)

  Other schemes are passed to the extending scheme handler (if
  one has been registered). This way, requests such as http:...
  can be processed.

  The life cycle of a DataLine:
    Upon construction, no URI is attached yet.
    Call open() to associate a URI.
    Repeatedly call save() or get() to retrieve data.
    Call close() to close the resource.
    Call the destructor.

  The 'write' data line with the scheme of 'arg' will need to be
  accessible to the user even after the Processor object is destroyed;
  it is then freed by 'SablotFreeBuffer'.
*****************************************************************/

/*****************************************************************
DataLine::DataLine()

  This constructor just sets everything to zeroes and such.
*****************************************************************/

DataLine::DataLine()
{
    mode = DLMODE_NONE;       
    scheme = URI_NONE;      
    f = NULL;               
    buffer = NULL;          
    outBuf = NULL;
    bufCurr = 0;
    fileIsStd = FALSE;
    utf16Encoded = FALSE;
    handler = NULL;
    handle = 0;
}

/*****************************************************************
DataLine::~DataLine()

  The destructor asserts that the data line had been closed.
*****************************************************************/

DataLine::~DataLine()
{
    assert(mode == DLMODE_CLOSED);
    assert(!f);
    // if there is an outBuf, delete it now
    if (outBuf)
        delete outBuf;
}

/*****************************************************************
DataLine::open()

  Opens the data line for a given URI and access mode. Actual
  data transfer is only done on subsequent get() or save() calls.
  open() tries to call the extending scheme handler if it cannot
  handle a request itself.

ARGS
_uri        the URI identifier for the resource, including the
            scheme (e.g. "file:///x.xml")
_baseUri    the base URI used in case the reference in _uri is
            relative
_mode       the access mode (DLMODE_READ, DLMODE_WRITE)
*****************************************************************/

eFlag DataLine::open(const char *_uri, DLAccessMode _mode)
{
    assert(mode == DLMODE_NONE);  // the buffer must not be open yet
    mode = _mode;
    fullUri = (char*)_uri;

    // combine _uri and _baseUri into one
    Str strScheme, strPath;
    scheme = uri2SchemePath(_uri, strScheme, strPath);

    char *name = (char*) strPath;
    switch(scheme)
    {
    case URI_FILE:
        {
            if (name[0] == '/' && name[1] == '/')
                name += 2;          // skipping the "//" in front
            // try to open the file
            if (!(f = stdopen(name,
                _mode == DLMODE_WRITE ? "wt" : "rt")))
                return NOT_OK;
            // set fileIsStd if filename is "stdin", "stdout" or "stderr"
            fileIsStd = isstd(name);
        }; break;
    case URI_ARG:
        {
            // if opening for read access, get the pointer to the argument contents
            // plus some extra information
            if (_mode == DLMODE_READ)
            {
                Bool isUTF16 = FALSE;
                /* 
                  call getArg to set:
                  'buffer' to the text block pointer
                  'isUTF16' to TRUE if the arg is UTF-16 encoded
                */
                E( proc -> getArg(name, buffer, isUTF16) );
            }
            // if opening for write access, just allocate a new dynamic block
            else
                outBuf = new DynBlock;
        }; break;
    default:
        {
            // try the extending scheme handler
            // ask the handler address from the Processor
            handler = proc -> getSchemeHandler();
            // if there is no handler, report unsupported scheme
            if (!handler)
                Err1(E1_UNSUPPORTED_SCHEME, strScheme);
            // call the handler's open() function, obtaining a handle
            switch(handler -> open((char*) strScheme, name, &handle))
            {
            case SH_ERR_UNSUPPORTED_SCHEME:     // scheme not supported
                Err1(E1_UNSUPPORTED_SCHEME, strScheme);
            case SH_ERR_NOT_OK:
                Err1(E1_URI_OPEN, strScheme + ":" + strPath);
            };
        };
    };
    // open successfully completed. Set the new mode.
    mode = _mode;
    return OK;
}

/*****************************************************************
DataLine::close()

  closes the resource attached to this data line.
*****************************************************************/
eFlag DataLine::close()
{
    assert(mode != DLMODE_NONE);
    switch(scheme)
    {
    case URI_FILE:
        {
            assert(f);
            if (!fileIsStd)
            {
                if (fclose(f))
                    Err1(E1_URI_CLOSE, fullUri);
            };
            f = NULL;
        }; break;
    case URI_ARG:
        break;
    case URI_EXTENSION:
        {
            if(NZ(handler) -> close(handle))
                Err1(E1_URI_CLOSE, fullUri);
        }; break;
    };
    mode = DLMODE_CLOSED;
    return OK;
}

/*****************************************************************
save()

  saves an UTF-8 string pointed to by data to the data line.
  This is the place to perform any recoding, escaping and other operations
  that require char-by-char scanning of the string.
*****************************************************************/

int my_wcslen(const char *p)
{
    int len;
    for (len = 2;  *(short int*)p; p += 2, len += 2);
    return len;
}

eFlag DataLine::save(const char *data)
{
    assert(mode == DLMODE_WRITE); // assume the file open for writing
    // PRB: data should be output one character at a time, so far we are
    // outputting in bulk
    int length = utf16Encoded ? my_wcslen(data) : strlen(data);
    switch (scheme)             // choose the output procedure 
    {
    case URI_FILE:              // file: scheme
        {
            assert(f);          // the file must be open
            fputs(data, f);
        }; break;
    case URI_ARG:               // arg: scheme
        {
            assert(outBuf);     // the output buffer must exist
            outBuf -> nadd(data, length); 
        }; break;
    case URI_EXTENSION:         // external handler
        {
            int actual = length;
            if( NZ(handler) -> put(handle, data, &actual) )
                Err1(E1_URI_WRITE, fullUri);
        };
    }
    return OK;
}

/*................................................................
pointsAtEnd()

  DESCRIPTION
a macro that returns nonzero if the given char* points at a 
string terminator

  ARGS
p       the pointer
is16    TRUE iff the string is UTF-16
................................................................*/

#define pointsAtEnd(p, is16) ((is16) ? (!*(unsigned short*)(p)) : (!*(p)))

/*****************************************************************
    get()

- retrieves at most 'maxcount' bytes into buffer 'dest'.
- input need not be NULL-terminated
- if a terminating 0 is reached, copying stops
*****************************************************************/

int DataLine::get(char *dest,int maxcount)
{
    int result;
    assert(mode == DLMODE_READ);  // assume the file open for reading
    switch(scheme)
    {
    case URI_FILE:
        {
            assert(f);          // the file must be open
            result = fread(dest,1,maxcount,f);
            // return the number of bytes read
        }; break;
    case URI_ARG:
        {
            assert(buffer);     // the buffer must exist
            // do a 'strncpy' that shifts dest and bufCurr;
            // i counts the number of bytes transferred
	    int i;
            for (i = 0; 
                (!pointsAtEnd(buffer + bufCurr, utf16Encoded)) && (i < maxcount); 
                i++)
                {
                    *(dest++) = buffer[bufCurr++];
                };
                result = i;
        }; break;
    case URI_EXTENSION:         // external handler
        {
            int actual = maxcount;
            if( NZ(handler) -> get(handle, dest, &actual) )
                Err1(E1_URI_READ, fullUri);
            result = actual;
        }; break;
    }
    return result;              // return the number of bytes read
}

/*****************************************************************
getOutBuffer()

  returns the pointer to the output buffer which may be used after
  all processing is finished (remains allocated along with the
  whole DataLine)
*****************************************************************/

DynBlock* DataLine::getOutBuffer()
{
    // check that the output buffer exists and that we're open for write
    assert(mode == DLMODE_WRITE && scheme == URI_ARG); 
    return NZ(outBuf); // -> getPointer();
}
