/*
 * Sablotron - XSL processor
 * Copyright (C) 2000 Ginger Alliance Ltd.
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
 * GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

/*****************************************************************
    uri.h
    Sablotron XSLT processor project
    
This module does the work related to retrieving data in answer to
a URI request.
*****************************************************************/

#ifndef UriHIncl
#define UriHIncl

#include "base.h"
#include "shandler.h"

/*  URIScheme
    possible URI schemes for a DataLine 
        - file
        - arg for access to named buffers
        - other scheme to by handled by an extension (a SchemeHandler)
*/
typedef enum
{
    URI_FILE, URI_ARG, URI_EXTENSION, URI_NONE
} URIScheme;

/*****************************************************************
makeAbsoluteURI()

  merges a (possibly relative) URI reference with a base URI.
ARGS
  uri       the URI reference
  base      the base URI
RETURNS
  absolute  the result
*****************************************************************/


URIScheme makeAbsoluteURI(const char* uri ,const char* base, Str& absolute);


/*  DLAccessMode
    possible access modes for a DataLine
*/

typedef enum
{
    DLMODE_NONE, DLMODE_READ, DLMODE_WRITE, DLMODE_CLOSED
} DLAccessMode;

/*****************************************************************
DynBlock

  is a class for "dynamic memory blocks", i.e. blocks that can 
  grow in size.
*****************************************************************/

struct DynBlockItem
{
    char *data;
    int byteCount;
    DynBlockItem *next;
};

class DynBlock
{
public:
    DynBlock();
    ~DynBlock();
    void nadd(const char *data, int bytes);
    char *getPointer();
    int byteCount;
private:
    void compact();
    DynBlockItem *first, *last;
};

/*****************************************************************  
DataLine
    a class associated to any data source or destination at a given URI; 
    - can be open for reading or for writing
    - 'Processor' class has a list of open DataLines 
        together with the associated trees (serving as a cache)
*****************************************************************/

class DataLine
{
public:
    DataLine();
    ~DataLine();
    // opens the resource at _uri for reading/writing based on _mode:
    eFlag open(const char *_uri, DLAccessMode _mode);
    eFlag close();
    // sends data to the resource
    eFlag save(const char *data);
    int get(char *where, int maxcount);
    DynBlock *getOutBuffer();
    Str fullUri;
    DLAccessMode mode;
    URIScheme scheme;
private:
    FILE *f;
    char *buffer;
    DynBlock *outBuf;
    int bufCurr;
    SchemeHandler *handler;
    int handle;
    Bool 
        fileIsStd,
        utf16Encoded;
};
    
#endif
