/*
 * Sablotron - XSL processor
 * Copyright (C) 2000 Ginger Alliance Ltd.
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
 * GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include "vars.h"
#include "situa.h"

#define iterate(VAR, LST) for(int VAR = 0; VAR < (LST).number(); VAR++)

void varDump(VarsList *which, char *text)
{
    printf("'%s': variable dump at level %d/%d\n", text, which -> currCallLevel, 
        which -> currNestLevel);
    for (int i = 0; i < which -> number(); i++)
    {
        VarBindings *record = (*which)[i];
        printf("%s ",(char *)(record -> varname.local));
        for (int j = 0; j < record -> bindings.number(); j++)
        {
            printf("%s%d/%d ", (record -> bindings[j] -> prebinding ? "P":""),
                record -> bindings[j] -> callLevel, 
                record -> bindings[j] -> nestLevel);
        }
        puts("");
    }
    puts("");
}

/*****************************************************************
    V a r B i n d i n g I t e m
*****************************************************************/

VarBindingItem::~VarBindingItem()
{
    delete expr;
}

/*****************************************************************
    V a r B i n d i n g s
*****************************************************************/

VarBindings::VarBindings()
{
}

VarBindings::~VarBindings()
{
    bindings.freeall(FALSE);
}

/*****************************************************************
    V a r s L i s t
*****************************************************************/

VarsList::VarsList()
: SList<VarBindings*>(1)
{
    currCallLevel = currNestLevel = 0;
}

VarsList::~VarsList()
{
    freeall(FALSE);
}

VarBindings *VarsList::find(QName &name)
{
    for (int i = 0; i < number(); i++)
        if ((*this)[i] -> varname == name)
            return (*this)[i];
    return NULL;
}

int VarsList::compare(int first, int second)
{
    assert((first >= 0) && (first < number()));
    assert((second >= 0) && (second < number()));
    QName 
        &q1 = (*this)[first] -> varname,
        &q2 = (*this)[second] -> varname;
    int u = q1.uri.compare(q2.uri);
    if (u) return u;
    u = q1.local.compare(q2.local);
    return u;
}

VarBindings *VarsList::getOrAdd(QName &q)
{
    VarBindings *record = find(q);
    if (!record)
    {
        record = new VarBindings;
        record -> varname = q;
        insert(record);
    };
    return record;
}

eFlag VarsList::addBinding(QName& q, Expression *e, Bool force)
{
    VarBindings *record = getOrAdd(q);
    VarBindingItem *lastbind = NULL;

    if (!record -> bindings.isEmpty())
    {
        lastbind = record -> bindings.last();
        if (lastbind -> callLevel == currCallLevel) // && !lastbind -> prebinding)
            Err(ET_MULT_ASSIGNMENT);
    }

    /*
        construct the new binding item. if there is a pre-binding that applies,
        just change it to a binding and drop the current expression. otherwise,
        add the new item.
    */

    VarBindingItem *it;
    record -> bindings.append(it = new VarBindingItem);
    it -> callLevel = currCallLevel;
    it -> nestLevel = currNestLevel;
    it -> prebinding = FALSE;
    if (lastbind && lastbind -> prebinding && 
        lastbind -> callLevel == (currCallLevel - 1) && 
        lastbind -> nestLevel == currNestLevel &&
        !force)
    {
        it -> expr = new Expression(e -> ownerV);
        E( lastbind -> expr -> eval(*(it -> expr), NULL) );
        delete e;
    }
    else
        it -> expr = e;
    return OK;
}

eFlag VarsList::addPrebinding(QName &q, Expression *e)
{
    VarBindings *record = getOrAdd(q);
    VarBindingItem *lastbind = NULL;

    if (!record -> bindings.isEmpty())
    {
        lastbind = record -> bindings.last();
        if (lastbind -> nestLevel == currNestLevel && 
            lastbind -> callLevel == currCallLevel &&    // new
            lastbind -> prebinding)
            Err(ET_MULT_ASSIGNMENT);
    }

    VarBindingItem *it = new VarBindingItem;
    it -> callLevel = currCallLevel; 
    it -> nestLevel = currNestLevel;
    it -> expr = e;
    it -> prebinding = TRUE;
    record -> bindings.append(it);

    return OK;
}

void VarsList::rmBinding(QName &q)
{
    VarBindings *record = find(q);
    assert(record && record -> bindings.number());
    record -> bindings.freelast(FALSE);
}

Expression *VarsList::getBinding(QName &q)
{
    VarBindings *record = find(q);
    if (!record || record -> bindings.isEmpty()) 
        return NULL;

    for (int i = record -> bindings.number() - 1; 
        (i >= 0) && (record -> bindings[i] -> callLevel == currCallLevel);
        i--)
        {
            if (!record -> bindings[i] -> prebinding)
                return record -> bindings[i] -> expr;
        };

    // try to return the global value
//    if (!record -> bindings[0] -> callLevel && !record -> bindings[0] -> prebinding)
    if (record -> bindings[0] -> callLevel == 1)
        return record -> bindings[0] -> expr;
    else 
    {
        if ((record -> bindings.number() > 1) && 
        (record -> bindings[1] -> callLevel == 1))
            return record -> bindings[1] -> expr;
        else
            return NULL;
    }
}

void VarsList::startCall()
{
    currCallLevel++;
    // varDump(this, "startCall");
}

/*================================================================
endCall
================================================================*/

void VarsList::_endCall(Bool rmLastLevelPrebs)
{
//    varDump(this, "_endCall before");
    VarBindingItem* lastbind;
    for (int i = 0; i < number(); i++)
    {
        VarBindings *record = (*this)[i];
        if (!record -> bindings.isEmpty())
        {
            while ( !record -> bindings.isEmpty() &&
                    (((lastbind=record -> bindings.last()) -> 
                            callLevel == currCallLevel) ||
                        (rmLastLevelPrebs &&
                            lastbind -> callLevel == currCallLevel - 1 &&
                            lastbind -> prebinding)) && 
                    lastbind -> nestLevel >= currNestLevel) // was ==
            {
                record -> bindings.freelast(FALSE);
            }
        }
    }
//    varDump(this, "_endCall after");
    currCallLevel--;
}

void VarsList::endCall()
{
    _endCall(TRUE);
}

void VarsList::startNested()
{
    currNestLevel++;
}

void VarsList::endNested()
{
    currNestLevel--;
}

void VarsList::startApplyOne()
{
    startCall();
}

void VarsList::endApplyOne()
{
    _endCall(FALSE);
}

void VarsList::rmPrebindings()
{
    VarBindingItem* lastbind;
    for (int i = 0; i < number(); i++)
    {
        VarBindings *record = (*this)[i];
        if (!record -> bindings.isEmpty() && 
            (lastbind = record -> bindings.last()) -> callLevel == currCallLevel &&
            lastbind -> prebinding && 
            lastbind -> nestLevel >= currNestLevel)
        {
            record -> bindings.freelast(FALSE);
        }
    }
}

