/*
 * Sablotron - XSL processor
 * Copyright (C) 2000 Ginger Alliance Ltd.
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
 * GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#ifndef VertsHIncl
#define VertsHIncl

#include "base.h"
#include "datastr.h"

#include "expr.h"

class Context;
class Expression;

class AttList;

#define toNS(v) (cast(NmSpace*,v))
#define toE(v) (cast(Element*,v))
#define toX(v) (cast(XSLElement*,v))

/****************************************
V e r t e x
****************************************/

class Vertex /* : public SabObj */
{
public:
    Vertex(VTYPE avt = VT_VERTEX_WF)
        : vt(avt), parent(NULL), ordinal(0)
    {
        stamp = lineno = 0;
    };
    virtual ~Vertex();
    virtual eFlag execute(Context *);
   	virtual eFlag value(Str &, Context *);
    virtual void
        copy(Tree *, Bool);
    void output(DataLine &);
    virtual void speak(Str &, SpeakMode);
    void setParent(Vertex *v);
    const QName& getName() const;
    Vertex *parent;
    int 
        ordinal,
        lineno;
    VTYPE vt;
    QName name;
    int 
        stamp;          // the vertex number in doc order (0 = root)
};

/****************************************
V e r t e x L i s t
****************************************/

class VertexList: public PList<Vertex*>
{
public:
    VertexList();
    ~VertexList();
    eFlag execute (Context*);
    eFlag value(Str&, Context *);
    virtual void speak(Str &, SpeakMode);
    virtual void append(Vertex *);
    int strip();
    void copy(Tree *, Bool);
};

/****************************************
D a d d y
****************************************/

class Daddy : public Vertex
{
public:
    Daddy(VTYPE avt = VT_DADDY_WF)
        : Vertex(avt)
    {};
    virtual ~Daddy(void);
    virtual eFlag execute(Context *c);
   	virtual eFlag value(Str &, Context *);
    virtual eFlag newChild(Vertex*);
    virtual eFlag checkChildren();
    virtual void speak(Str &, SpeakMode);
    virtual int strip();
    //
    VertexList contents;
};

/****************************************
R o o t N o d e
****************************************/

class RootNode : public Daddy
{
public:
    RootNode()
        : Daddy(VT_ROOT_WF)
    {
    };
    virtual ~RootNode();
    virtual eFlag execute(Context *c);
    virtual eFlag newChild(Vertex*);
    virtual eFlag checkChildren();
    virtual void speak(Str &, SpeakMode);
    virtual void
        copy(Tree *, Bool);
    //
    VertexList prolog;
};

/****************************************
A t t r i b u t e
****************************************/

class Attribute : public Vertex
{
public:
    Attribute(const QName&, const Str&, XSL_ATT);
    virtual ~Attribute();
    eFlag buildExpr(Bool, ExType);
    virtual eFlag execute(Context *c);
    virtual void speak(Str &, SpeakMode);
   	virtual eFlag value(Str &, Context *);
    virtual void
        copy(Tree *, Bool);
    Str cont;
    Expression *expr;
    XSL_ATT op;
};

/****************************************
A t t L i s t
****************************************/

class AttList: public VertexList
{
public:
    Attribute *find(XSL_ATT);
};

/*****************************************************************
    N m S p a c e
*****************************************************************/

class NmSpace: public Vertex
{
public:
    NmSpace(const Str&, const Str&);
    virtual ~NmSpace();
    Str
        prefix,
        uri;
    virtual eFlag execute(Context *);
    virtual void speak(Str &, SpeakMode);
   	virtual eFlag value(Str &, Context *);
    virtual void
        copy(Tree *, Bool);
};

/*****************************************************************
    N S L i s t
*****************************************************************/

class NSList : public VertexList
{
public:
    ~NSList();
    NmSpace *find(Str&) const;
    eFlag resolve(Str&, Bool);
    void unresolve(Str&) const;
//    NSList& operator= (const NSList&);
    void giveCurrent(NSList &, Tree*) const;
};

/****************************************
E l e m e n t
****************************************/

class Element : public Daddy
{
public:
    Element(QName&, Tree *_ownerT, VTYPE = VT_ELEMENT_WF);
    virtual ~Element();
    virtual eFlag execute(Context *c);
    virtual eFlag newChild(Vertex*);
    virtual void speak(Str &, SpeakMode);
    void removeBindings();
    virtual void
        copy(Tree *, Bool);
    //
    NSList
        namespaces;
    AttList 
        atts;
    Tree
        *ownerT;
};

/****************************************
T e x t
****************************************/

class Text : public Vertex
{
public:
    Str cont;
    Text(char *acont,int alen=0)
        : Vertex(VT_TEXT_WF)
    {
        if (alen)
            cont.nset(acont,alen);
        else cont = acont;
    };
    virtual ~Text();
    virtual eFlag execute(Context *);
    virtual void speak(Str &, SpeakMode);
   	virtual eFlag value(Str &, Context *);
    virtual void
        copy(Tree *, Bool);
    Bool allWhite();
};

/****************************************
X S L E l e m e n t
****************************************/

class XSLElement : public Element
{
public:
	eFlag checkAtts();       
    XSLElement(QName&, Tree *_ownerT, XSL_OP);
    virtual eFlag execute(Context *c);
    virtual eFlag newChild(Vertex*);
    virtual eFlag checkChildren();
    void checkExtraChildren(int& k);
    eFlag processChildren();
    Expression *getAttExpr(XSL_ATT);
    virtual int strip();
    XSL_OP op;
    VertexList defs;
};
#endif //ifndef VertsHIncl

